#include "pvOptVtk.h"

#include <QDebug>
#include <QStringList>

#include <vtkSmartPointer.h>
#include <vtkCubeSource.h>
#include <vtkSphereSource.h>
#include <vtkActor.h>
#include <vtkPolyData.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkRenderWindow.h>
#include <vtkRenderWindowInteractor.h>
#include <vtkRenderer.h>
#include <vtkBorderWidget.h>
#include <vtkBorderRepresentation.h>
#include <vtkFeatureEdges.h>
#include <vtkAssembly.h>
#include <vtkNew.h>
#include <vtkSTLReader.h>
#include <vtkOpenFOAMReader.h>
#include <vtkNamedColors.h>
#include <vtkAxesActor.h>
#include <vtkOrientationMarkerWidget.h>
#include <vtkInteractorStyleMultiTouchCamera.h>

//-----------------------------------------------------------------------------

pvOptVtk::pvOptVtk() 
{
}

void pvOptVtk::visualizeBoundBox(std::vector<double> bounds, QStringList surfacePaths)
{
  // Create the RenderWindow, Renderer and both Actors
  vtkSmartPointer<vtkRenderer> ren1 = vtkSmartPointer<vtkRenderer>::New();
  vtkSmartPointer<vtkRenderWindow> renWin = vtkSmartPointer<vtkRenderWindow>::New();
  renWin->AddRenderer(ren1);
  renWin->SetWindowName("PvOptUMesh Background Mesh Visualization");

  vtkSmartPointer<vtkRenderWindowInteractor> iren = vtkSmartPointer<vtkRenderWindowInteractor>::New();
  vtkSmartPointer<vtkInteractorStyleMultiTouchCamera> istyle = vtkSmartPointer<vtkInteractorStyleMultiTouchCamera>::New();
  iren->SetInteractorStyle(istyle);
  iren->SetRenderWindow(renWin);

  vtkSmartPointer<vtkNamedColors> namedColors = vtkSmartPointer<vtkNamedColors>::New();

  // Create Pipeline
  vtkSmartPointer<vtkCubeSource> box = vtkSmartPointer<vtkCubeSource>::New();
  box->SetBounds(bounds.data());
  
  vtkSmartPointer<vtkAssembly> assembly = vtkSmartPointer<vtkAssembly>::New();
  for (int i = 0; i < surfacePaths.size(); i++)
  {
    vtkSmartPointer<vtkSTLReader> reader = vtkSmartPointer<vtkSTLReader>::New();
    reader->SetFileName(surfacePaths[i].toStdString().c_str());
    reader->Update();
    vtkSmartPointer<vtkPolyDataMapper> stlMapper = vtkSmartPointer<vtkPolyDataMapper>::New();
    stlMapper->SetInputConnection(reader->GetOutputPort());

    vtkSmartPointer<vtkActor> stlActor = vtkSmartPointer<vtkActor>::New();
    stlActor->SetMapper(stlMapper);
    stlActor->GetProperty()->SetDiffuseColor(namedColors->GetColor3d("LightSteelBlue").GetData());
    assembly->AddPart(stlActor);
  }
  
  vtkSmartPointer<vtkFeatureEdges> edges = vtkSmartPointer<vtkFeatureEdges>::New();
  edges->SetInputConnection(box->GetOutputPort());
  vtkSmartPointer<vtkPolyDataMapper> mapper = vtkSmartPointer<vtkPolyDataMapper>::New();
  mapper->SetInputConnection(edges->GetOutputPort());
  vtkSmartPointer<vtkActor> boxActor = vtkSmartPointer<vtkActor>::New();
  boxActor->SetMapper(mapper);
  boxActor->GetProperty()->SetColor(namedColors->GetColor3d("Tomato").GetData());

  // Create the widget and its representation
  vtkSmartPointer<vtkBorderRepresentation> rep = vtkSmartPointer<vtkBorderRepresentation>::New();
  rep->ProportionalResizeOn();
  rep->SetShowBorderToOn();

  vtkSmartPointer<vtkBorderWidget> widget = vtkSmartPointer<vtkBorderWidget>::New();
  widget->SetInteractor(iren);
  widget->SetRepresentation(rep);
  widget->SelectableOff();

  vtkSmartPointer<vtkAxesActor> axes = vtkSmartPointer<vtkAxesActor>::New();
  vtkSmartPointer<vtkOrientationMarkerWidget> coords = vtkSmartPointer<vtkOrientationMarkerWidget>::New();
  double rgba[4]{0.0, 0.0, 0.0, 0.0};
  namedColors->GetColor("Carrot", rgba);
  coords->SetOutlineColor(rgba[0], rgba[1], rgba[2]);
  coords->SetOrientationMarker(axes);
  coords->SetInteractor(iren);
  coords->SetViewport(0.0, 0.0, 0.4, 0.4);
  coords->SetEnabled(1);
  coords->InteractiveOn();

  ren1->AddActor(boxActor);
  ren1->AddActor(assembly);
  ren1->SetBackground(0.1, 0.2, 0.4);
  renWin->SetSize(600, 600);

  iren->Initialize();
  renWin->Render();
  widget->On();
  iren->Start();
}

void pvOptVtk::visualizeFFD(std::vector<std::vector<double>> ffdPoints, std::string foamPath)
{
  // Create the RenderWindow, Renderer and both Actors
  vtkSmartPointer<vtkRenderer> ren1 = vtkSmartPointer<vtkRenderer>::New();
  vtkSmartPointer<vtkRenderWindow> renWin = vtkSmartPointer<vtkRenderWindow>::New();
  renWin->AddRenderer(ren1);
  renWin->SetWindowName("FFD Point Visualization");

  vtkSmartPointer<vtkRenderWindowInteractor> iren = vtkSmartPointer<vtkRenderWindowInteractor>::New();
  vtkSmartPointer<vtkInteractorStyleMultiTouchCamera> istyle = vtkSmartPointer<vtkInteractorStyleMultiTouchCamera>::New();
  iren->SetInteractorStyle(istyle);
  iren->SetRenderWindow(renWin);

  vtkSmartPointer<vtkNamedColors> namedColors = vtkSmartPointer<vtkNamedColors>::New();

  // Create Pipeline
  // OpenFOAMReader
  vtkSmartPointer<vtkOpenFOAMReader> reader = vtkSmartPointer<vtkOpenFOAMReader>::New();
  reader->SetFileName(foamPath.c_str());
  reader->Update();
  vtkSmartPointer<vtkPolyDataMapper> meshMapper = vtkSmartPointer<vtkPolyDataMapper>::New();
  meshMapper->SetInputConnection(reader->GetOutputPort());
  vtkSmartPointer<vtkActor> meshActor = vtkSmartPointer<vtkActor>::New();
  meshActor->SetMapper(meshMapper);
  meshActor->GetProperty()->SetDiffuseColor(namedColors->GetColor3d("LightSteelBlue").GetData());
  
  // FFD Points
  vtkSmartPointer<vtkAssembly> ffdAssembly = vtkSmartPointer<vtkAssembly>::New();
  for (int i = 0; i<ffdPoints.size(); i++)
  {
    std::vector<double> loc = ffdPoints[i];
    vtkSmartPointer<vtkSphereSource> sphereSource = vtkSmartPointer<vtkSphereSource>::New();
    sphereSource->SetCenter(loc[0],loc[1],loc[2]); //set location at ffd point
    sphereSource->SetRadius(0.1); //set radius in relation to spread of ffd points
    sphereSource->SetPhiResolution(100);
    sphereSource->SetThetaResolution(100);
    vtkSmartPointer<vtkPolyDataMapper> pointMapper = vtkSmartPointer<vtkPolyDataMapper>::New();
    pointMapper->SetInputConnection(sphereSource->GetOutputPort());
    vtkSmartPointer<vtkActor> pointActor = vtkSmartPointer<vtkActor>::New();
    pointActor->SetMapper(pointMapper);
    pointActor->GetProperty()->SetColor(namedColors->GetColor3d("Tomato").GetData());
    ffdAssembly->AddPart(pointActor);
  }

  // Create the widget and its representation
  vtkSmartPointer<vtkBorderRepresentation> rep = vtkSmartPointer<vtkBorderRepresentation>::New();
  rep->ProportionalResizeOn();
  rep->SetShowBorderToOn();

  vtkSmartPointer<vtkBorderWidget> widget = vtkSmartPointer<vtkBorderWidget>::New();
  widget->SetInteractor(iren);
  widget->SetRepresentation(rep);
  widget->SelectableOff();

  vtkSmartPointer<vtkAxesActor> axes = vtkSmartPointer<vtkAxesActor>::New();
  vtkSmartPointer<vtkOrientationMarkerWidget> coords = vtkSmartPointer<vtkOrientationMarkerWidget>::New();
  double rgba[4]{0.0, 0.0, 0.0, 0.0};
  namedColors->GetColor("Carrot", rgba);
  coords->SetOutlineColor(rgba[0], rgba[1], rgba[2]);
  coords->SetOrientationMarker(axes);
  coords->SetInteractor(iren);
  coords->SetViewport(0.0, 0.0, 0.4, 0.4);
  coords->SetEnabled(1);
  coords->InteractiveOn();

  ren1->AddActor(meshActor);
  ren1->AddActor(ffdAssembly);
  ren1->SetBackground(0.1, 0.2, 0.4);
  renWin->SetSize(600, 600);

  iren->Initialize();
  renWin->Render();
  widget->On();
  iren->Start();
}

/*void pvOptVtk::executePythonCmd(std::string path)
{
  vtkSmartPointer<vtkPythonInterpreter> pythonManager = vtkSmartPointer<vtkPythonInterpreter>::New();
  if (pythonManager->RunSimpleString(path.c_str()) == -1)
  {
    qDebug() << "error in python script";
  }
}*/

/*void pvOptVtk::refreshFoamReader(std::string rootPath)
{
  std::ostringstream reloadPath;
  reloadPath << rootPath << "/reloadFoam.py";
  std::ofstream reloadPy(reloadPath.str().c_str(), std::ios::out);
  
  reloadPy<<
  "from paraview.simple import *\n"
  "\n"
  "allSources = GetSources()\n"
  "\n"
  "for s in allSources:\n"
  "    if 'foam' in s[0]:\n"
  "        foam = allSources[s]\n"
  "        foam.Refresh()\n"
  "        foam.UpdatePipeline()\n"
  "        foam.UpdatePipelineInformation()\n"
  "\n"
  "view = GetActiveView()" << std::endl;

  reloadPy.flush();
  reloadPy.close();

  executePythonCmd(reloadPath.str());

  QFile file(reloadPath.str().c_str());
  file.remove();
}*/